@@ -21,7 +21,7 @@
 #include <stdlib.h>
 #include <glib/gi18n.h>
 #include <gio/gio.h>
-#include <gtk/gtkicontheme.h>
+#include <gtk/gtk.h>
 
 #include "vinagre-connection.h"
 #include "vinagre-enums.h"
@@ -33,8 +33,10 @@ struct _VinagreConnectionPrivate
   gchar *host;
   gint   port;
   gchar *name;
+  gchar *username;
   gchar *password;
   gchar *desktop_name;
+  gint   shared;
   gboolean view_only;
   gboolean scaling;
   gboolean fullscreen;
@@ -47,13 +49,15 @@ enum
   PROP_HOST,
   PROP_PORT,
   PROP_NAME,
+  PROP_USERNAME,
   PROP_PASSWORD,
   PROP_DESKTOP_NAME,
   PROP_BEST_NAME,
   PROP_ICON,
   PROP_VIEW_ONLY,
   PROP_SCALING,
-  PROP_FULLSCREEN
+  PROP_FULLSCREEN,
+  PROP_SHARED,
 };
 
 gint   vinagre_connection_default_port [VINAGRE_CONNECTION_PROTOCOL_INVALID-1] = {5900, 3389};
@@ -71,11 +75,13 @@ vinagre_connection_init (VinagreConnection *conn)
   conn->priv->host = NULL;
   conn->priv->port = 0;
   conn->priv->password = NULL;
+  conn->priv->username = NULL;
   conn->priv->name = NULL;
   conn->priv->desktop_name = NULL;
   conn->priv->view_only = FALSE;
   conn->priv->scaling = FALSE;
   conn->priv->fullscreen = FALSE;
+  conn->priv->shared = -1;
 }
 
 static void
@@ -83,21 +89,11 @@ vinagre_connection_finalize (GObject *object)
 {
   VinagreConnection *conn = VINAGRE_CONNECTION (object);
 
-  if (conn->priv->host)
-    g_free (conn->priv->host);
-  conn->priv->host = NULL;
-
-  if (conn->priv->password)
-    g_free (conn->priv->password);
-  conn->priv->password = NULL;
-
-  if (conn->priv->name)
-    g_free (conn->priv->name);
-  conn->priv->name = NULL;
-
-  if (conn->priv->desktop_name)
-    g_free (conn->priv->desktop_name);
-  conn->priv->desktop_name = NULL;
+  g_free (conn->priv->host);
+  g_free (conn->priv->username);
+  g_free (conn->priv->password);
+  g_free (conn->priv->name);
+  g_free (conn->priv->desktop_name);
 
   G_OBJECT_CLASS (vinagre_connection_parent_class)->finalize (object);
 }
@@ -125,6 +121,10 @@ vinagre_connection_set_property (GObject *object, guint prop_id, const GValue *v
 	vinagre_connection_set_port (conn, g_value_get_int (value));
 	break;
 
+      case PROP_USERNAME:
+	vinagre_connection_set_username (conn, g_value_get_string (value));
+	break;
+
       case PROP_PASSWORD:
 	vinagre_connection_set_password (conn, g_value_get_string (value));
 	break;
@@ -149,6 +149,10 @@ vinagre_connection_set_property (GObject *object, guint prop_id, const GValue *v
 	vinagre_connection_set_fullscreen (conn, g_value_get_boolean (value));
 	break;
 
+      case PROP_SHARED:
+	vinagre_connection_set_shared (conn, g_value_get_int (value));
+	break;
+
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	break;
@@ -179,6 +183,10 @@ vinagre_connection_get_property (GObject *object, guint prop_id, GValue *value,
 	g_value_set_int (value, conn->priv->port);
 	break;
 
+      case PROP_USERNAME:
+	g_value_set_string (value, conn->priv->username);
+	break;
+
       case PROP_PASSWORD:
 	g_value_set_string (value, conn->priv->password);
 	break;
@@ -211,6 +219,10 @@ vinagre_connection_get_property (GObject *object, guint prop_id, GValue *value,
 	g_value_set_boolean (value, conn->priv->fullscreen);
 	break;
 
+      case PROP_SHARED:
+	g_value_set_int (value, conn->priv->shared);
+	break;
+
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	break;
@@ -268,6 +280,18 @@ vinagre_connection_class_init (VinagreConnectionClass *klass)
                                                       G_PARAM_STATIC_NAME |
                                                       G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (object_class,
+                                   PROP_NAME,
+                                   g_param_spec_string ("username",
+                                                        "username",
+	                                                "username (if any) necessary for complete this connection",
+                                                        NULL,
+	                                                G_PARAM_READWRITE |
+                                                        G_PARAM_CONSTRUCT |
+                                                        G_PARAM_STATIC_NICK |
+                                                        G_PARAM_STATIC_NAME |
+                                                        G_PARAM_STATIC_BLURB));
+
   g_object_class_install_property (object_class,
                                    PROP_PASSWORD,
                                    g_param_spec_string ("password",
@@ -359,6 +383,20 @@ vinagre_connection_class_init (VinagreConnectionClass *klass)
                                                         G_PARAM_STATIC_NAME |
                                                         G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (object_class,
+                                   PROP_PORT,
+                                   g_param_spec_int ("shared",
+                                                     "shared flag",
+	                                              "if the server should allow more than one client connected",
+                                                      -1,
+                                                      1,
+                                                      -1,
+	                                              G_PARAM_READWRITE |
+                                                      G_PARAM_CONSTRUCT |
+                                                      G_PARAM_STATIC_NICK |
+                                                      G_PARAM_STATIC_NAME |
+                                                      G_PARAM_STATIC_BLURB));
+
 }
 
 VinagreConnection *
@@ -418,14 +456,30 @@ vinagre_connection_get_port (VinagreConnection *conn)
   return conn->priv->port;
 }
 
+void
+vinagre_connection_set_username (VinagreConnection *conn,
+			     const gchar *username)
+{
+  g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
+
+  g_free (conn->priv->username);
+  conn->priv->username = g_strdup (username);
+}
+const gchar *
+vinagre_connection_get_username (VinagreConnection *conn)
+{
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), NULL);
+
+  return conn->priv->username;
+}
+
 void
 vinagre_connection_set_password (VinagreConnection *conn,
 			         const gchar *password)
 {
   g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
 
-  if (conn->priv->password)
-    g_free (conn->priv->password);
+  g_free (conn->priv->password);
   conn->priv->password = g_strdup (password);
 }
 const gchar *
@@ -442,8 +496,7 @@ vinagre_connection_set_name (VinagreConnection *conn,
 {
   g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
 
-  if (conn->priv->name)
-    g_free (conn->priv->name);
+  g_free (conn->priv->name);
   conn->priv->name = g_strdup (name);
 }
 const gchar *
@@ -460,8 +513,7 @@ vinagre_connection_set_desktop_name (VinagreConnection *conn,
 {
   g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
 
-  if (conn->priv->desktop_name)
-    g_free (conn->priv->desktop_name);
+  g_free (conn->priv->desktop_name);
   conn->priv->desktop_name = g_strdup (desktop_name);
 }
 const gchar *
@@ -489,6 +541,23 @@ vinagre_connection_get_best_name (VinagreConnection *conn)
   return NULL;
 }
 
+void
+vinagre_connection_set_shared (VinagreConnection *conn,
+			       gint value)
+{
+  g_return_if_fail (VINAGRE_IS_CONNECTION (conn));
+  g_return_if_fail (value >=0 && value <=1);
+
+  conn->priv->shared = value;
+}
+gint
+vinagre_connection_get_shared (VinagreConnection *conn)
+{
+  g_return_val_if_fail (VINAGRE_IS_CONNECTION (conn), -1);
+
+  return conn->priv->shared;
+}
+
 VinagreConnection *
 vinagre_connection_clone (VinagreConnection *conn)
 {
@@ -498,6 +567,7 @@ vinagre_connection_clone (VinagreConnection *conn)
 
   vinagre_connection_set_host (new_conn, vinagre_connection_get_host (conn));
   vinagre_connection_set_port (new_conn, vinagre_connection_get_port (conn));
+  vinagre_connection_set_username (new_conn, vinagre_connection_get_username (conn));
   vinagre_connection_set_password (new_conn, vinagre_connection_get_password (conn));
   vinagre_connection_set_name (new_conn, vinagre_connection_get_name (conn));
   vinagre_connection_set_desktop_name (new_conn, vinagre_connection_get_desktop_name (conn));
@@ -563,18 +633,19 @@ vinagre_connection_split_string (const gchar *uri,
 }
 
 VinagreConnection *
-vinagre_connection_new_from_string (const gchar *uri, gchar **error_msg)
+vinagre_connection_new_from_string (const gchar *uri, gchar **error_msg, gboolean use_bookmarks)
 {
-  VinagreConnection *conn;
+  VinagreConnection *conn = NULL;
   gint    port;
   gchar  *host;
 
   if (!vinagre_connection_split_string (uri, &host, &port, error_msg))
     return NULL;
 
-  conn = vinagre_bookmarks_exists (vinagre_bookmarks_get_default (),
-                                   host,
-                                   port);
+  if (use_bookmarks)
+    conn = vinagre_bookmarks_exists (vinagre_bookmarks_get_default (),
+				     host,
+				     port);
   if (!conn)
     {
       conn = vinagre_connection_new ();
@@ -587,19 +658,24 @@ vinagre_connection_new_from_string (const gchar *uri, gchar **error_msg)
 }
 
 VinagreConnection *
-vinagre_connection_new_from_file (const gchar *uri, gchar **error_msg)
+vinagre_connection_new_from_file (const gchar *uri, gchar **error_msg, gboolean use_bookmarks)
 {
   GKeyFile          *file;
-  GError            *error = NULL;
+  GError            *error;
   gboolean           loaded;
-  VinagreConnection *conn = NULL;
-  gchar             *host = NULL;
+  VinagreConnection *conn;
+  gchar             *host, *actual_host, *data;
   gint               port;
   int                file_size;
-  char              *data = NULL;
   GFile             *file_a;
 
   *error_msg = NULL;
+  host = NULL;
+  data = NULL;
+  conn = NULL;
+  error = NULL;
+  file = NULL;
+  conn = NULL;
 
   file_a = g_file_new_for_commandline_arg (uri);
   loaded = g_file_load_contents (file_a,
@@ -615,52 +691,84 @@ vinagre_connection_new_from_file (const gchar *uri, gchar **error_msg)
 	  *error_msg = g_strdup (error->message);
 	  g_error_free (error);
 	}
+      else
+	*error_msg = g_strdup (_("Could not open the file."));
 
-      if (data)
-	g_free (data);
-
-      g_object_unref (file_a);
-      return NULL;
+      goto the_end;
     }
 
   file = g_key_file_new ();
   loaded = g_key_file_load_from_data (file,
 				      data,
 				      file_size,
-				      G_KEY_FILE_NONE,
+				      0,
 				      &error);
-  if (loaded)
-    {
-      host = g_key_file_get_string (file, "connection", "host", NULL);
-      port = g_key_file_get_integer (file, "connection", "port", NULL);
-      if (host)
-	{
-	  conn = vinagre_bookmarks_exists (vinagre_bookmarks_get_default (),
-                                           host,
-                                           port);
-	  if (!conn)
-	    {
-	      conn = vinagre_connection_new ();
-	      vinagre_connection_set_host (conn, host);
-	      vinagre_connection_set_port (conn, port);
-	    }
-	  g_free (host);
-	}
-    }
-  else
+  if (!loaded)
     {
       if (error)
 	{
 	  *error_msg = g_strdup (error->message);
 	  g_error_free (error);
 	}
+      else
+	*error_msg = g_strdup (_("Could not parse the file."));
+
+      goto the_end;
     }
 
-  if (data)
-    g_free (data);
+  host = g_key_file_get_string (file, "connection", "host", NULL);
+  port = g_key_file_get_integer (file, "connection", "port", NULL);
+  if (host)
+    {
+      if (!port)
+	{
+	  if (!vinagre_connection_split_string (host, &actual_host, &port, error_msg))
+	    goto the_end;
 
-  g_key_file_free (file);
+	  g_free (host);
+	  host = actual_host;
+	}
+
+      if (use_bookmarks)
+        conn = vinagre_bookmarks_exists (vinagre_bookmarks_get_default (), host, port);
+      if (!conn)
+	{
+	  gchar *username, *password;
+	  gint shared;
+	  GError *e = NULL;
+
+	  conn = vinagre_connection_new ();
+	  vinagre_connection_set_host (conn, host);
+	  vinagre_connection_set_port (conn, port);
+
+	  username = g_key_file_get_string  (file, "connection", "username", NULL);
+	  vinagre_connection_set_username (conn, username);
+	  g_free (username);
+
+	  password = g_key_file_get_string  (file, "connection", "password", NULL);
+	  vinagre_connection_set_password (conn, password);
+	  g_free (password);
+
+	  shared = g_key_file_get_integer (file, "options", "shared", &e);
+	  if (e)
+	    g_error_free (e);
+	  else
+	    if (shared == 0 || shared == 1)
+	      vinagre_connection_set_shared (conn, shared);
+	    else
+	      g_message (_("Bad value for 'shared' flag: %d. It is supposed to be 0 or 1. Ignoring it."), shared);
+	}
+
+      g_free (host);
+    }
+  else
+    *error_msg = g_strdup (_("Could not find the host address in the file."));
+
+the_end:
+  g_free (data);
   g_object_unref (file_a);
+  if (file)
+    g_key_file_free (file);
 
   return conn;
 }
@@ -743,12 +851,12 @@ vinagre_connection_get_string_rep (VinagreConnection *conn,
   if (has_protocol)
     {
       uri = g_string_new (vinagre_connection_protos [conn->priv->protocol-1]);
-      g_string_append_printf (uri, "%s", "://");
+      g_string_append (uri, "://");
     }
   else
     uri = g_string_new (NULL);
 
-  g_string_append_printf (uri, "%s", conn->priv->host);
+  g_string_append (uri, conn->priv->host);
 
   if (vinagre_connection_default_port [conn->priv->protocol-1] != conn->priv->port)
     g_string_append_printf (uri, "::%d", conn->priv->port);
@@ -759,4 +867,4 @@ vinagre_connection_get_string_rep (VinagreConnection *conn,
   return result;
 }
 
-/* vim: ts=8 */
+/* vim: set ts=8: */
