@@ -43,7 +43,7 @@ struct _VinagreTabPrivate
   VinagreConnection *conn;
   VinagreNotebook   *nb;
   VinagreWindow     *window;
-  gboolean           save_password;
+  gboolean           save_credential;
   guint32            keyring_item_id;
   VinagreTabState    state;
   gchar             *clipboard_str;
@@ -71,6 +71,8 @@ enum
   PROP_0,
   PROP_CONN,
   PROP_WINDOW,
+  PROP_ORIGINAL_WIDTH,
+  PROP_ORIGINAL_HEIGHT
 };
 
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -140,6 +142,12 @@ vinagre_tab_get_property (GObject    *object,
       case PROP_WINDOW:
         g_value_set_object (value, tab->priv->window);
 	break;
+      case PROP_ORIGINAL_WIDTH:
+        g_value_set_int (value, vinagre_tab_get_original_width (tab));
+	break;
+      case PROP_ORIGINAL_HEIGHT:
+        g_value_set_int (value, vinagre_tab_get_original_height (tab));
+	break;
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	break;			
@@ -178,24 +186,36 @@ vinagre_tab_finalize (GObject *object)
 {
   VinagreTab *tab = VINAGRE_TAB (object);
 
-  g_signal_handlers_disconnect_by_func (tab->priv->window,
-  					vinagre_tab_window_state_cb,
-  					tab);
-  g_object_unref (tab->priv->conn);
-  tab->priv->conn = NULL;
-
   g_free (tab->priv->clipboard_str);
-  tab->priv->clipboard_str = NULL;
 
   G_OBJECT_CLASS (vinagre_tab_parent_class)->finalize (object);
 }
 
+static void
+vinagre_tab_dispose (GObject *object)
+{
+  VinagreTab *tab = VINAGRE_TAB (object);
+
+  if (tab->priv->conn)
+    {
+      g_signal_handlers_disconnect_by_func (tab->priv->window,
+  					    vinagre_tab_window_state_cb,
+  					    tab);
+
+      g_object_unref (tab->priv->conn);
+      tab->priv->conn = NULL;
+    }
+
+  G_OBJECT_CLASS (vinagre_tab_parent_class)->dispose (object);
+}
+
 static void 
 vinagre_tab_class_init (VinagreTabClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
   object_class->finalize = vinagre_tab_finalize;
+  object_class->dispose  = vinagre_tab_dispose;
   object_class->get_property = vinagre_tab_get_property;
   object_class->set_property = vinagre_tab_set_property;
 
@@ -222,6 +242,28 @@ vinagre_tab_class_init (VinagreTabClass *klass)
 							G_PARAM_STATIC_NICK |
 							G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (object_class,
+				   PROP_ORIGINAL_WIDTH,
+				   g_param_spec_int ("original-width",
+						     "Original width",
+						     "The original width of the remote screen",
+						     -1, G_MAXINT, 0,
+						      G_PARAM_READABLE |
+						      G_PARAM_STATIC_NAME |
+						      G_PARAM_STATIC_NICK |
+						      G_PARAM_STATIC_BLURB));
+
+  g_object_class_install_property (object_class,
+				   PROP_ORIGINAL_HEIGHT,
+				   g_param_spec_int ("original-height",
+						     "Original height",
+						     "The original height of the remote screen",
+						     -1, G_MAXINT, 0,
+						      G_PARAM_READABLE |
+						      G_PARAM_STATIC_NAME |
+						      G_PARAM_STATIC_NICK |
+						      G_PARAM_STATIC_BLURB));
+
   signals[TAB_CONNECTED] =
 		g_signal_new ("tab-connected",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -259,16 +301,25 @@ static void
 open_vnc (VinagreTab *tab)
 {
   gchar *port;
+  gint  shared;
 
   vnc_display_set_force_size (VNC_DISPLAY(tab->priv->vnc),
 			      !vinagre_connection_get_scaling (tab->priv->conn));
 
   port = g_strdup_printf ("%d", vinagre_connection_get_port (tab->priv->conn));
 
+  shared = vinagre_connection_get_shared (tab->priv->conn);
+  if (shared == -1)
+    g_object_get (vinagre_prefs_get_default (),
+		  "shared-flag", &shared,
+		  NULL);
+  vnc_display_set_shared_flag (VNC_DISPLAY (tab->priv->vnc),
+			       shared);
+
   if (vnc_display_open_host (VNC_DISPLAY(tab->priv->vnc), vinagre_connection_get_host (tab->priv->conn), port))
     gtk_widget_grab_focus (tab->priv->vnc);
   else
-    vinagre_utils_show_error (_("Error connecting to host."), NULL);
+    vinagre_utils_show_error (NULL, _("Error connecting to host."), NULL);
 
   g_free (port);
 }
@@ -300,13 +351,13 @@ vnc_auth_failed_cb (VncDisplay *vnc, const gchar *msg, VinagreTab *tab)
   message = g_string_new (NULL);
   name = vinagre_connection_get_best_name (vinagre_tab_get_conn (tab));
 
-  g_string_printf (message, _("Authentication to host \"%s\" has failed"),
+  g_string_printf (message, _("Authentication to host <i>%s</i> has failed"),
 		   name);
   if (msg)
   	g_string_append_printf (message, " (%s)", msg);
-  g_string_append_printf (message, ".");
+  g_string_append_c (message, '.');
 
-  vinagre_utils_show_error (message->str, GTK_WINDOW (tab->priv->window));
+  vinagre_utils_show_error (_("Authentication failed"), message->str, GTK_WINDOW (tab->priv->window));
   g_string_free (message, TRUE);
   g_free (name);
 
@@ -328,11 +379,11 @@ vnc_auth_unsupported_cb (VncDisplay *vnc, guint auth_type, VinagreTab *tab)
   message = g_string_new (NULL);
   name = vinagre_connection_get_best_name (vinagre_tab_get_conn (tab));
 
-  g_string_printf (message, _("Authentication method to host \"%s\" is unsupported. (%u)"),
+  g_string_printf (message, _("Authentication method to host <i>%s</i> is unsupported. (%u)"),
 		   name,
 		   auth_type);
 
-  vinagre_utils_show_error (message->str, GTK_WINDOW (tab->priv->window));
+  vinagre_utils_show_error (_("Authentication unsupported"), message->str, GTK_WINDOW (tab->priv->window));
   g_string_free (message, TRUE);
   g_free (name);
 
@@ -409,7 +460,8 @@ vinagre_tab_add_recent_used (VinagreTab *tab)
   data->is_private = FALSE;
 
   if (!gtk_recent_manager_add_full (manager, uri->str, data))
-    vinagre_utils_show_error (_("Error saving recent connection."),
+    vinagre_utils_show_error (NULL,
+			      _("Error saving recent connection."),
 			      GTK_WINDOW (tab->priv->window));
 
   g_string_free (uri, TRUE);
@@ -420,41 +472,43 @@ vinagre_tab_add_recent_used (VinagreTab *tab)
 }
 
 static void
-vinagre_tab_save_password (VinagreTab *tab)
+vinagre_tab_save_credential (VinagreTab *tab)
 {
   GnomeKeyringResult result;
 
-  if (!tab->priv->save_password)
+  if (!tab->priv->save_credential)
     return;
 
   result = gnome_keyring_set_network_password_sync (
-                NULL,           /* default keyring */
-                NULL,           /* user            */
-                NULL,           /* domain          */
-                vinagre_connection_get_host (tab->priv->conn),   /* server          */
-                NULL,           /* object          */
-                "rfb",          /* protocol        */
-                "vnc-password", /* authtype        */
-                vinagre_connection_get_port (tab->priv->conn),           /* port            */
-                vinagre_connection_get_password (tab->priv->conn),       /* password        */
+                NULL,                                              /* default keyring */
+                vinagre_connection_get_username (tab->priv->conn), /* user            */
+                NULL,                                              /* domain          */
+                vinagre_connection_get_host (tab->priv->conn),     /* server          */
+                NULL,                                              /* object          */
+                "rfb",                                             /* protocol        */
+                "vnc-password",                                    /* authtype        */
+                vinagre_connection_get_port (tab->priv->conn),     /* port            */
+                vinagre_connection_get_password (tab->priv->conn), /* password        */
                 &tab->priv->keyring_item_id);
 
   if (result != GNOME_KEYRING_RESULT_OK)
-    vinagre_utils_show_error (_("Error saving the password on the keyring."),
+    vinagre_utils_show_error (NULL,
+			      _("Error saving the credential on the keyring."),
 			      GTK_WINDOW (tab->priv->window));
 
-  tab->priv->save_password = FALSE;
+  tab->priv->save_credential = FALSE;
 }
 
-static gchar *
-vinagre_tab_find_password (VinagreTab *tab)
+static gboolean
+vinagre_tab_find_credentials (VinagreTab *tab, gchar **username, gchar **password)
 {
   GnomeKeyringNetworkPasswordData *found_item;
   GnomeKeyringResult               result;
   GList                           *matches;
-  gchar                           *password;
   
-  matches = NULL;
+  matches   = NULL;
+  *username = NULL;
+  *password = NULL;
 
   result = gnome_keyring_find_network_password_sync (
                 NULL,           /* user     */
@@ -463,20 +517,22 @@ vinagre_tab_find_password (VinagreTab *tab)
 		NULL,           /* object   */
 		"rfb",          /* protocol */
 		"vnc-password", /* authtype */
-		vinagre_connection_get_port (tab->priv->conn),           /* port     */
+		vinagre_connection_get_port (tab->priv->conn),   /* port     */
 		&matches);
 
   if (result != GNOME_KEYRING_RESULT_OK || matches == NULL || matches->data == NULL)
-    return NULL;
+    return FALSE;
 
   found_item = (GnomeKeyringNetworkPasswordData *) matches->data;
 
-  password = g_strdup (found_item->password);
+  *username = g_strdup (found_item->user);
+  *password = g_strdup (found_item->password);
+  
   tab->priv->keyring_item_id = found_item->item_id;
 
   gnome_keyring_network_password_list_free (matches);
 
-  return password;
+  return TRUE;
 }
 
 static void
@@ -511,7 +567,7 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreTab *tab)
   g_free (name);
 
   vinagre_window_set_title (tab->priv->window);
-  vinagre_tab_save_password (tab);
+  vinagre_tab_save_credential (tab);
   vinagre_tab_add_recent_used (tab);
 
   tab->priv->state = VINAGRE_TAB_STATE_CONNECTED;
@@ -523,76 +579,136 @@ vnc_initialized_cb (VncDisplay *vnc, VinagreTab *tab)
 		 0);
 }
 
-static gchar *
-ask_password(VinagreTab *tab)
+static gboolean
+ask_credential (VinagreTab *tab,
+		gboolean    need_username,
+		gboolean    need_password,
+		gchar     **username,
+		gchar     **password)
 {
   GladeXML   *xml;
   const char *glade_file;
-  GtkWidget  *password_dialog, *password_entry, *host_label, *save_password_check;
-  gchar      *password = NULL, *name;
+  GtkWidget  *password_dialog, *password_entry, *host_label, *save_credential_check;
+  GtkWidget  *password_label, *username_label, *username_entry;
+  gchar      *name, *label;
   int         result;
 
+  *username = NULL;
+  *password = NULL;
+
   glade_file = vinagre_utils_get_glade_filename ();
   xml = glade_xml_new (glade_file, NULL, NULL);
 
-  password_dialog = glade_xml_get_widget (xml, "password_required_dialog");
+  password_dialog = glade_xml_get_widget (xml, "auth_required_dialog");
   gtk_window_set_transient_for (GTK_WINDOW(password_dialog), GTK_WINDOW(tab->priv->window));
 
   host_label = glade_xml_get_widget (xml, "host_label");
   name = vinagre_connection_get_best_name (tab->priv->conn);
-  gtk_label_set_text (GTK_LABEL (host_label), name);
+  label = g_strdup_printf ("<i>%s</i>", name);
+  gtk_label_set_markup (GTK_LABEL (host_label), label);
   g_free (name);
+  g_free (label);
+
+  password_entry = glade_xml_get_widget (xml, "password_entry");
+  username_entry = glade_xml_get_widget (xml, "username_entry");
+  password_label = glade_xml_get_widget (xml, "password_label");
+  username_label = glade_xml_get_widget (xml, "username_label");
+  save_credential_check = glade_xml_get_widget (xml, "save_credential_check");
+
+  if (!need_username)
+    {
+      gtk_widget_hide (username_label);
+      gtk_widget_hide (username_entry);
+    }
+
+  if (!need_password)
+    {
+      gtk_widget_hide (password_label);
+      gtk_widget_hide (password_entry);
+    }
 
   result = gtk_dialog_run (GTK_DIALOG (password_dialog));
   if (result == -5)
     {
-      password_entry = glade_xml_get_widget (xml, "password_entry");
-      password = g_strdup (gtk_entry_get_text (GTK_ENTRY (password_entry)));
-
-      save_password_check = glade_xml_get_widget (xml, "save_password_check");
-      tab->priv->save_password = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (save_password_check));
+      *username = g_strdup (gtk_entry_get_text (GTK_ENTRY (username_entry)));
+      *password = g_strdup (gtk_entry_get_text (GTK_ENTRY (password_entry)));
+      tab->priv->save_credential = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (save_credential_check));
     }
 
   gtk_widget_destroy (GTK_WIDGET (password_dialog));
   g_object_unref (xml);
 
-  return password;
+  return result == -5;
 }
 
 static void
 vnc_authentication_cb (VncDisplay *vnc, GValueArray *credList, VinagreTab *tab)
 {
-  gchar *password;
+  gchar *username, *password;
+  gboolean need_password, need_username;
   int i;
 
   if (credList == NULL)
     return;
 
+  need_password = FALSE;
+  need_username = FALSE;
+  username = NULL;
+  password = NULL;
+
   for (i = 0; i < credList->n_values; i++) {
     switch (g_value_get_enum (&credList->values[i]))
       {
-        case VNC_DISPLAY_CREDENTIAL_PASSWORD:
-          password = vinagre_tab_find_password (tab);
-          if (!password)
-            {
-              password = ask_password (tab);
-              if (!password) {
-                vinagre_notebook_remove_tab (tab->priv->nb, tab);
-                return;
-              }
-            }
+	case VNC_DISPLAY_CREDENTIAL_USERNAME:
+	  if (vinagre_connection_get_username (tab->priv->conn))
+	    {
+	      vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_USERNAME, vinagre_connection_get_username (tab->priv->conn));
+	      break;
+	    }
+	  need_username= TRUE;
+	  break;
 
-          vinagre_connection_set_password (tab->priv->conn, password);
-          vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
-
-          g_free (password);
-          break;
+	case VNC_DISPLAY_CREDENTIAL_PASSWORD:
+	  if (vinagre_connection_get_password (tab->priv->conn))
+	    {
+	      vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, vinagre_connection_get_password (tab->priv->conn));
+	      break;
+	    }
+	  need_password = TRUE;
+	  break;
 
         case VNC_DISPLAY_CREDENTIAL_CLIENTNAME:
           vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_CLIENTNAME, "vinagre");
           break;
       }
   }
+
+  if (need_password || need_username)
+    {
+      if (!vinagre_tab_find_credentials (tab, &username, &password))
+	{
+	  if (!ask_credential (tab, need_username, need_password, &username, &password))
+	    {
+	      vinagre_notebook_remove_tab (tab->priv->nb, tab);
+	      return;
+	    }
+	}
+
+      if (need_username && username)
+	{
+	  vinagre_connection_set_username (tab->priv->conn, username);
+	  vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_USERNAME, username);
+	}
+
+      if (need_password && password)
+	{
+	  vinagre_connection_set_password (tab->priv->conn, password);
+	  vnc_display_set_credential (vnc, VNC_DISPLAY_CREDENTIAL_PASSWORD, password);
+	}
+
+      g_free (username);
+      g_free (password);
+    }
 }
 
 static void
@@ -609,11 +725,19 @@ vnc_pointer_ungrab_cb (VncDisplay *vnc, VinagreTab *tab)
   vinagre_window_set_title (tab->priv->window);
 }
 
-static void vnc_bell_cb (VncDisplay *vnc, VinagreTab *tab)
+static void
+vnc_bell_cb (VncDisplay *vnc, VinagreTab *tab)
 {
   gdk_window_beep (GTK_WIDGET (tab->priv->window)->window);
 }
 
+static void
+vnc_desktop_resize_cb (VncDisplay *vnc, int x, int y, VinagreTab *tab)
+{
+  g_object_notify (G_OBJECT (tab), "original-width");
+  g_object_notify (G_OBJECT (tab), "original-height");
+}
+
 static void
 close_button_clicked (GtkToolButton *button,
 		      VinagreTab    *tab)
@@ -621,6 +745,13 @@ close_button_clicked (GtkToolButton *button,
   vinagre_notebook_remove_tab (tab->priv->nb, tab);
 }
 
+static void
+minimize_button_clicked (GtkToolButton *button,
+			 VinagreTab    *tab)
+{
+  gtk_window_iconify (GTK_WINDOW (tab->priv->window));
+}
+
 static void
 fullscreen_button_clicked (GtkToolButton *button,
 			   VinagreTab    *tab)
@@ -637,7 +768,7 @@ screenshot_button_clicked (GtkToolButton *button,
 
 static void
 cad_button_clicked (GtkToolButton *button,
-			   VinagreTab    *tab)
+		    VinagreTab    *tab)
 {
   vinagre_tab_send_ctrlaltdel (tab);
 }
@@ -677,16 +808,8 @@ setup_layout (VinagreTab *tab)
   tab->priv->toolbar = gtk_toolbar_new ();
   GTK_WIDGET_SET_FLAGS (tab->priv->toolbar, GTK_NO_SHOW_ALL);
 
-  gtk_toolbar_set_tooltips (GTK_TOOLBAR (tab->priv->toolbar), TRUE);
   gtk_toolbar_set_style (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOLBAR_BOTH_HORIZ);
 
-  /* Leave fullscreen */
-  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_LEAVE_FULLSCREEN));
-  gtk_widget_show (GTK_WIDGET (button));
-  gtk_tool_item_set_is_important (GTK_TOOL_ITEM (button), TRUE);
-  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
-  g_signal_connect (button, "clicked", G_CALLBACK (fullscreen_button_clicked), tab);
-
   /* Close connection */
   button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_CLOSE));
   gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Close connection"));
@@ -694,6 +817,20 @@ setup_layout (VinagreTab *tab)
   gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
   g_signal_connect (button, "clicked", G_CALLBACK (close_button_clicked), tab);
 
+  /* Leave fullscreen */
+  button = GTK_WIDGET (gtk_tool_button_new_from_stock (GTK_STOCK_LEAVE_FULLSCREEN));
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (fullscreen_button_clicked), tab);
+
+  /* Minimize window */
+  button = GTK_WIDGET (gtk_tool_button_new (NULL, NULL));
+  gtk_tool_item_set_tooltip_text (GTK_TOOL_ITEM (button), _("Minimize window"));
+  gtk_tool_button_set_icon_name (GTK_TOOL_BUTTON (button), "view-minimize");
+  gtk_widget_show (GTK_WIDGET (button));
+  gtk_toolbar_insert (GTK_TOOLBAR (tab->priv->toolbar), GTK_TOOL_ITEM (button), 0);
+  g_signal_connect (button, "clicked", G_CALLBACK (minimize_button_clicked), tab);
+
   /* Space */
   button = GTK_WIDGET (gtk_separator_tool_item_new ());
   gtk_separator_tool_item_set_draw (GTK_SEPARATOR_TOOL_ITEM (button), FALSE);
@@ -748,11 +885,10 @@ setup_layout (VinagreTab *tab)
 static void
 vinagre_tab_init (VinagreTab *tab)
 {
-  GtkWidget     *viewport;
-  gboolean       shared;
+  GtkWidget *viewport;
 
   tab->priv = VINAGRE_TAB_GET_PRIVATE (tab);
-  tab->priv->save_password = FALSE;
+  tab->priv->save_credential = FALSE;
   tab->priv->keyring_item_id = 0;
   tab->priv->state = VINAGRE_TAB_STATE_INITIALIZING;
   tab->priv->clipboard_str = NULL;
@@ -774,12 +910,6 @@ vinagre_tab_init (VinagreTab *tab)
   viewport = gtk_bin_get_child (GTK_BIN (tab->priv->scroll));
   gtk_viewport_set_shadow_type(GTK_VIEWPORT (viewport), GTK_SHADOW_NONE);
 
-  g_object_get (vinagre_prefs_get_default (),
-		"shared-flag", &shared,
-		NULL);
-  vnc_display_set_shared_flag (VNC_DISPLAY (tab->priv->vnc),
-			       shared);
-
   g_signal_connect (tab->priv->vnc,
 		    "vnc-connected",
 		    G_CALLBACK (vnc_connected_cb),
@@ -830,6 +960,11 @@ vinagre_tab_init (VinagreTab *tab)
 		    G_CALLBACK (vnc_bell_cb),
 		    tab);
 
+  g_signal_connect (tab->priv->vnc,
+		    "vnc-desktop-resize",
+		    G_CALLBACK (vnc_desktop_resize_cb),
+		    tab);
+
   setup_layout (tab);
 
   gtk_box_pack_end (GTK_BOX(tab), tab->priv->layout, TRUE, TRUE, 0);
@@ -956,7 +1091,7 @@ vinagre_tab_take_screenshot (VinagreTab *tab)
   }
 
   gtk_widget_destroy (dialog);
-  gdk_pixbuf_unref (pix);
+  g_object_unref (pix);
   g_string_free (suggested_filename, TRUE);
   g_free (name);
 }
@@ -994,18 +1129,10 @@ vinagre_tab_set_scaling (VinagreTab *tab, gboolean active) {
   if (vnc_display_get_scaling (VNC_DISPLAY (tab->priv->vnc)) == active)
     return TRUE;
 
-  if (active &&
-      gdk_screen_is_composited (gtk_widget_get_screen (GTK_WIDGET (tab->priv->window))))
-    {
-      vinagre_utils_show_error (_("Scaling does not work properly on composited windows. Disable the visual effects and try again."),
-				GTK_WINDOW (tab->priv->window));
-      return FALSE;
-    }
-
   vnc_display_set_force_size (VNC_DISPLAY(tab->priv->vnc), !active);
   if (!vnc_display_set_scaling (VNC_DISPLAY (tab->priv->vnc), active))
     {
-      vinagre_utils_show_error (_("Scaling is not supported on this installation.\n\nRead the README file (shipped with Vinagre) in order to know how to enable this feature."),
+      vinagre_utils_show_error (NULL, _("Scaling is not supported on this installation.\n\nRead the README file (shipped with Vinagre) in order to know how to enable this feature."),
 				GTK_WINDOW (tab->priv->window));
       return FALSE;
     }
@@ -1025,7 +1152,7 @@ vinagre_tab_set_scaling (VinagreTab *tab, gboolean active) {
 
 gboolean
 vinagre_tab_get_scaling (VinagreTab *tab) {
-  g_return_if_fail (VINAGRE_IS_TAB (tab));
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), FALSE);
 
   return vnc_display_get_scaling (VNC_DISPLAY (tab->priv->vnc));
 }
@@ -1042,7 +1169,7 @@ vinagre_tab_set_readonly (VinagreTab *tab, gboolean active) {
 
 gboolean
 vinagre_tab_get_readonly (VinagreTab *tab) {
-  g_return_if_fail (VINAGRE_IS_TAB (tab));
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), FALSE);
 
   return vnc_display_get_read_only (VNC_DISPLAY (tab->priv->vnc));
 }
@@ -1050,7 +1177,7 @@ vinagre_tab_get_readonly (VinagreTab *tab) {
 VinagreTabState
 vinagre_tab_get_state (VinagreTab *tab)
 {
-  g_return_if_fail (VINAGRE_IS_TAB (tab));
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), VINAGRE_TAB_STATE_INVALID);
 
   return tab->priv->state;
 }
@@ -1068,10 +1195,90 @@ vinagre_tab_get_from_connection (VinagreConnection *conn)
 }
 
 gboolean
-vinagre_tab_is_pointer_grab (VinagreTab *tab) {
-  g_return_if_fail (VINAGRE_IS_TAB (tab));
+vinagre_tab_is_pointer_grab (VinagreTab *tab)
+{
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), FALSE);
 
   return tab->priv->pointer_grab;
 }
 
+gint
+vinagre_tab_get_original_height (VinagreTab *tab)
+{
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), -1);
+
+  if (VNC_IS_DISPLAY (tab->priv->vnc))
+    return vnc_display_get_height (VNC_DISPLAY (tab->priv->vnc));
+  else
+    return -1;
+}
+
+gint
+vinagre_tab_get_original_width (VinagreTab *tab)
+{
+  g_return_val_if_fail (VINAGRE_IS_TAB (tab), -1);
+
+  if (VNC_IS_DISPLAY (tab->priv->vnc))
+    return vnc_display_get_width (VNC_DISPLAY (tab->priv->vnc));
+  else
+    return -1;
+}
+
+
+typedef struct _VinagrePrefSize {
+  gint width, height;
+  gulong sig_id;
+} VinagrePrefSize;
+
+static gboolean
+cb_unset_size (gpointer data)
+{
+  GtkWidget *widget = data;
+
+  gtk_widget_queue_resize_no_redraw (widget);
+
+  return FALSE;
+}
+
+static void
+cb_set_preferred_size (GtkWidget *widget, GtkRequisition *req,
+		       gpointer data)
+{
+  VinagrePrefSize *size = data;
+
+  req->width = size->width;
+  req->height = size->height;
+
+  g_signal_handler_disconnect (widget, size->sig_id);
+  g_free (size);
+  g_idle_add (cb_unset_size, widget);
+}
+
+void
+vinagre_widget_set_preferred_size (GtkWidget *widget, gint width,
+				 gint height)
+{
+  VinagrePrefSize *size = g_new (VinagrePrefSize, 1);
+
+  size->width = width;
+  size->height = height;
+  size->sig_id = g_signal_connect (widget, "size-request",
+				   G_CALLBACK (cb_set_preferred_size),
+				   size);
+
+  gtk_widget_queue_resize (widget);
+}
+
+void
+vinagre_tab_original_size (VinagreTab *tab)
+{
+  g_return_if_fail (VINAGRE_IS_TAB (tab));
+
+  gtk_window_unmaximize (GTK_WINDOW (tab->priv->window));
+  gtk_window_resize (GTK_WINDOW (tab->priv->window), 1, 1);
+  vinagre_widget_set_preferred_size (GTK_WIDGET (tab),
+				     vinagre_tab_get_original_width (tab),
+				     vinagre_tab_get_original_height (tab));
+}
+
 /* vim: set ts=8: */
